"use strict";

var _interopRequireDefault = require("babel-runtime/helpers/interop-require-default")["default"];

var _interopRequireWildcard = require("babel-runtime/helpers/interop-require-wildcard")["default"];

exports.__esModule = true;
exports.UnaryExpression = UnaryExpression;
exports.DoExpression = DoExpression;
exports.ParenthesizedExpression = ParenthesizedExpression;
exports.UpdateExpression = UpdateExpression;
exports.ConditionalExpression = ConditionalExpression;
exports.NewExpression = NewExpression;
exports.SequenceExpression = SequenceExpression;
exports.ThisExpression = ThisExpression;
exports.Super = Super;
exports.Decorator = Decorator;
exports.CallExpression = CallExpression;
exports.YieldExpression = YieldExpression;
exports.AwaitExpression = AwaitExpression;
exports.EmptyStatement = EmptyStatement;
exports.ExpressionStatement = ExpressionStatement;
exports.AssignmentPattern = AssignmentPattern;
exports.AssignmentExpression = AssignmentExpression;
exports.BinaryExpression = BinaryExpression;
exports.LogicalExpression = LogicalExpression;
exports.BindExpression = BindExpression;
exports.MemberExpression = MemberExpression;
exports.MetaProperty = MetaProperty;

var _isInteger = require("is-integer");

var _isInteger2 = _interopRequireDefault(_isInteger);

var _lodashLangIsNumber = require("lodash/lang/isNumber");

var _lodashLangIsNumber2 = _interopRequireDefault(_lodashLangIsNumber);

var _lodashLangClone = require("lodash/lang/clone");

var _lodashLangClone2 = _interopRequireDefault(_lodashLangClone);

var _babelTypes = require("babel-types");

var t = _interopRequireWildcard(_babelTypes);

var _horchata = require("horchata");

var SCIENTIFIC_NOTATION = /e/i;

function UnaryExpression(node) {
  // operator can be any token type that has prefix: true
  // TODO: generic node lookup
  var s = _lodashLangClone2["default"](_horchata.Token.stateFromCode(node.operator === "!" ? "not" : node.operator));
  s.meta = { unary: true };
  this.push(s);
  this.print(node, "argument");
}

function DoExpression(node) {
  this.push("do");
  this.printBlock(node);
}

function ParenthesizedExpression(node) {
  this.push("(");
  this.print(node, "expression");
  this.push(")");
}

function UpdateExpression(node) {
  if (node.prefix) {
    this.push(node.operator);
    this.print(node, "argument");
  } else {
    this.print(node, "argument");
    this.push(node.operator);
  }
}

// TODO: ToggleExpression - see frappe "nice alternative to nested ternaries"

function ConditionalExpression(node) {
  this.push("if");
  // TODO: or if parent is an expression
  if (!node.parenthesizedExpression) this.push("!");
  this.print(node, "test");
  this.push("then");
  this.print(node, "consequent");
  this.push("else");
  this.print(node, "alternate");
}

function NewExpression(node) {
  this.push("new");
  this.print(node, "callee");
  if (!node.emptyArguments || node.arguments.length) this.printArguments(node);
}

function SequenceExpression(node) {
  this.printMultiple(node, "expressions", { separator: ";" });
}

function ThisExpression() {
  this.push("this");
}

function Super() {
  this.push("super");
}

function Decorator(node) {
  this.push("@");
  this.print(node, "expression");
  this.newline();
}

function CallExpression(node) {
  this.print(node, "callee");
  this.printArguments(node);
}

function YieldExpression(node) {
  this.push("yield");

  if (node.delegate) {
    this.push("*");
  }

  if (node.argument) {
    this.print(node, "argument");
  }
}

function AwaitExpression(node) {
  this.push("await");

  if (node.argument) {
    this.print(node, "argument");
  }
}

function EmptyStatement() {
  this.push("pass");
  this.newline();
}

function ExpressionStatement(node) {
  this.print(node, "expression");
  this.newline();
}

function AssignmentPattern(node) {
  this.print(node, "left");
  this.push("=");
  this.print(node, "right");
}

function AssignmentExpression(node) {
  this.print(node, "left");
  // can be any token that has {isAssign: true}
  this.push(node.operator);
  this.print(node, "right");
}

function BinaryExpression(node) {
  // ensure valid
  if (t.isFunctionExpression(node.left) && node.left.body.body.length === 0) {
    var extra = node.left.extra == null ? node.left.extra = {} : node.left.extra;
    if (extra && !extra.parenthesized) {
      extra.parenthesized = true;
      extra.autoParens = true;
    }
  }
  // all other binary operators
  this.print(node, "left");
  var operator = node.operator;
  if (!this.dialect["equality-symbols"]) {
    switch (operator) {
      case "===":
        operator = "is";break;
      case "!==":
        operator = "isnt";break;
      case "==":
        operator = "like";break;
      case "!=":
        operator = "unlike";break;
    }
  }
  this.push(operator);
  this.print(node, "right");
}

function LogicalExpression(node) {
  // ensure valid
  if (t.isFunctionExpression(node.left) && node.left.body.body.length === 0) {
    var extra = node.left.extra == null ? node.left.extra = {} : node.left.extra;
    if (extra && !extra.parenthesized) {
      extra.parenthesized = true;
      extra.autoParens = true;
    }
  }
  this.print(node, "left");
  this.push(node.operator == "||" ? "or" : "and");
  this.print(node, "right");
}

function BindExpression(node) {
  if (node.object != null) this.print(node, "object");
  this.push("::");
  this.print(node, "callee");
}

function MemberExpression(node) {
  this.print(node, "object");

  if (!node.computed && t.isMemberExpression(node.property)) {
    throw new TypeError("Got a MemberExpression for MemberExpression property");
  }

  var computed = node.computed;
  if (t.isLiteral(node.property) && _lodashLangIsNumber2["default"](node.property.value)) {
    computed = true;
  }

  if (computed) {
    this.push("[");
    this.print(node, "property");
    this.push("]");
  } else {
    if (t.isLiteral(node.object)) {
      // TODO: preserve original format
      var val = this._stringLiteral(node.object);
      if (_isInteger2["default"](+val) && !SCIENTIFIC_NOTATION.test(val) && !this.endsWith(".")) {
        this.push(".");
      }
    }

    this.push(".");
    this.print(node, "property");
  }
}

// like `new.target`

function MetaProperty(node) {
  this.print(node, "meta");
  this.push(".");
  this.print(node, "property");
}

// TODO: SoakExpression
// TODO: NullCoalesceExpression