
// simple prints will print an ast, without regard for cst tokens
// preserved print will print an ast, while trying to preserve whitespace
// and other formatting included in cst tokens

"use strict";

var _inherits = require("babel-runtime/helpers/inherits")["default"];

var _classCallCheck = require("babel-runtime/helpers/class-call-check")["default"];

var _Object$assign = require("babel-runtime/core-js/object/assign")["default"];

var _interopRequireDefault = require("babel-runtime/helpers/interop-require-default")["default"];

var _interopRequireWildcard = require("babel-runtime/helpers/interop-require-wildcard")["default"];

exports.__esModule = true;

var _lodashLangIsArray = require("lodash/lang/isArray");

var _lodashLangIsArray2 = _interopRequireDefault(_lodashLangIsArray);

var _babelTypes = require("babel-types");

var t = _interopRequireWildcard(_babelTypes);

var _tacoBuffer = require("./taco-buffer");

var _tacoBuffer2 = _interopRequireDefault(_tacoBuffer);

var _horchata = require("horchata");

var _printer2 = require("./_printer");

var _printer3 = _interopRequireDefault(_printer2);

var _generatorsTacoBase = require("./generators/taco/base");

var baseGenerators = _interopRequireWildcard(_generatorsTacoBase);

var _generatorsTacoClasses = require("./generators/taco/classes");

var classesGenerators = _interopRequireWildcard(_generatorsTacoClasses);

var _generatorsTacoExpressions = require("./generators/taco/expressions");

var expressionsGenerators = _interopRequireWildcard(_generatorsTacoExpressions);

var _generatorsTacoLiterals = require("./generators/taco/literals");

var literalsGenerators = _interopRequireWildcard(_generatorsTacoLiterals);

var _generatorsTacoMethods = require("./generators/taco/methods");

var methodsGenerators = _interopRequireWildcard(_generatorsTacoMethods);

var _generatorsTacoModules = require("./generators/taco/modules");

var modulesGenerators = _interopRequireWildcard(_generatorsTacoModules);

var _generatorsTacoStatements = require("./generators/taco/statements");

var statementsGenerators = _interopRequireWildcard(_generatorsTacoStatements);

var _generatorsTacoTemplateLiterals = require("./generators/taco/template-literals");

var templateLiteralsGenerators = _interopRequireWildcard(_generatorsTacoTemplateLiterals);

var blockCommentJs = _horchata.tokComments.blockCommentMeta['/*'],
    blockCommentTaco = _horchata.tokComments.blockCommentMeta['#*'];

function isParenthesized(node) {
  return node.extra != null && node.extra.parenthesized || node.parenthesizedExpression;
}

function canOmitParens(node, parent) {
  return t.isObjectExpression(node) || t.isFunctionExpression(node) && node.id === null && t.isExportDefaultDeclaration(parent) || false;
}

var TacoscriptPrinter = (function (_TacoscriptTokenBuffer) {
  _inherits(TacoscriptPrinter, _TacoscriptTokenBuffer);

  function TacoscriptPrinter(ast, opts, code) {
    _classCallCheck(this, TacoscriptPrinter);

    _TacoscriptTokenBuffer.call(this, opts, code);
    this.ast = ast;
    this.code = code;
    this._printedCommentStarts = {};
  }

  TacoscriptPrinter.prototype.tokenize = function tokenize() {
    var ast = this.ast;
    // prints the ast down into the buffer
    this._simplePrint(ast, null, {});
    this._finishPrint(ast, null, {});
  };

  TacoscriptPrinter.prototype.print = function print(parent, prop) {
    var opts = arguments.length <= 2 || arguments[2] === undefined ? {} : arguments[2];

    if (this.format.preserve && parent.tokenElements && parent.tokenElements.length) {
      this._preservedPrint(parent, prop, opts);
    } else {
      this._simplePrint(parent[prop], parent, opts);
    }
  };

  TacoscriptPrinter.prototype._simplePrint = function _simplePrint(node, parent, opts) {
    this._simpleStartPrint(node, parent, opts);
    if (!this[node.type]) {
      throw new Error("Cannot print node of type " + node.type);
    }
    this[node.type](node, parent, opts);
    this._finishSimplePrint(node, parent, opts);
  };

  TacoscriptPrinter.prototype._preservedPrint = function _preservedPrint(parent, prop, opts) {
    var node = parent[prop];
    this._startPreservedPrint(parent, prop, opts);
    if (!this[node.type]) {
      throw new Error("Cannot print node of type " + node.type);
    }
    this[node.type](node, parent, opts);
    this._finishPrint(node, parent, opts);
  };

  TacoscriptPrinter.prototype._startPrint = function _startPrint(parent, prop, opts) {
    if (this.format.preserve && parent.tokenElements && parent.tokenElements.length) {
      return this._startPreservedPrint(parent, prop, opts);
    } else {
      return this._simpleStartPrint(parent[prop], parent, opts);
    }
  };

  TacoscriptPrinter.prototype._simpleStartPrint = function _simpleStartPrint(node, parent, opts) {
    this.printLeadingComments(node, parent);

    // TODO: catchup newlines

    if (opts.before) {
      opts.before();
    }

    // push mapping start pseudo-token
    if (this.opts.sourceMaps) {
      this.push({ type: 'mappingMark', value: { loc: node.loc.start, pos: node.start } });
    }
    if (isParenthesized(node) && !canOmitParens(node, parent)) this.push("(");
  };

  TacoscriptPrinter.prototype._startPreservedPrint = function _startPreservedPrint(parent, prop, opts) {
    var node = parent[prop];
    throw new Error('Not Implemented');
    // print tokens between prev sibling and node
    // * preserve correct indentation syntax
    // * push mapping start pseudo-token before the first non-whitespace/comment token
    // * keep state to indicate if leading parens, etc. were printed
    // run opts.before
  };

  TacoscriptPrinter.prototype._finishPrint = function _finishPrint(node, parent, opts) {
    if (this.format.preserve && node.tokenElements && node.tokenElements.length) {
      throw new Error('Not Implemented');
      // print all remaining unprinted tokens
      // * preserve correct indentation syntax
      //   * make sure that, if the last token is a block, trailing tokens are indented
      //     appropriately
      // * push mapping end pseudo-token after the last non-whitespace/comment token
      // * keep state to indicate if trailing parens, etc. were printed
      // run opts.after
    } else {
        this._finishSimplePrint(node, parent, opts);
      }
    this.flush();
  };

  TacoscriptPrinter.prototype._finishSimplePrint = function _finishSimplePrint(node, parent, opts) {
    if (isParenthesized(node) && !canOmitParens(node, parent)) this.push(")");
    // push mapping end pseudo-token
    if (this.opts.sourceMaps) {
      this.push({ type: 'mappingMark', value: { loc: node.loc.end, pos: node.end } });
    }

    if (opts.after) {
      opts.after();
    }

    this.printTrailingComments(node);
  };

  TacoscriptPrinter.prototype.printMultiple = function printMultiple(parent, prop) {
    var opts = arguments.length <= 2 || arguments[2] === undefined ? {} : arguments[2];

    var nodes = parent[prop];
    if (!nodes || !nodes.length) {
      if (this.format.preserve && parent.tokenElements && parent.tokenElements.length) {
        // if a node of child: prop is found
        // * print tokens leading up to last node of child: prop
        throw new Error("Not Implemented");
      }
      return;
    }
    if (this.format.preserve && parent.tokenElements && parent.tokenElements.length) {
      throw new Error("Not Implemented");
      // Iterate through tokenElements
      // print tokens leading
      // TODO: use a Map
      // print tokens between first of list and preceding sibling
      // print first child
      // iterate through rest of children
      //   * print tokens between prev child and this child
      //   * print child
      //   * if a token element that isn't prop is found
      //     * if there are more token elements of child: prop in the list
      //       * throw invalid ast
      //     * else switch to simple printing
      // if more token elements of child: prop are in the list
      // * print the remaining tokens leading up to the last element of child: prop
    } else {
        this._simplePrintMultiple(nodes, parent, opts);
      }
  };

  TacoscriptPrinter.prototype._simplePrintMultiple = function _simplePrintMultiple(nodes, parent, opts) {
    var _this = this;

    var len = nodes.length;
    var separator = opts.separator ? _lodashLangIsArray2["default"](opts.separator) ? opts.separator : [opts.separator] : [];
    var separatorIsNewline = separator.length === 1 && (separator[0].type === 'newline' || separator[0].type === _horchata.tokTypes.newline);
    var node = undefined,
        i = undefined;

    if (opts.indent) {
      this.indent();
    }

    var after = separatorIsNewline ? function () {
      if (opts.iterator) opts.iterator(node, i);
      if (opts.separator && i < len - 1) {
        _this.newline();
      }
    } : function () {
      if (opts.iterator) opts.iterator(node, i);
      if (opts.separator && i < len - 1) {
        _this.push.apply(_this, separator);
      }
    };

    var printOpts = {
      statement: opts.statement,
      after: after
    };

    for (i = 0; i < len; i++) {
      node = nodes[i];
      if (node) {
        this._simplePrint(node, parent, printOpts);
      } else {
        // preserve holes, especially in arrays
        this.push("pass");
        after();
      }
    }

    if (opts.indent) this.dedent();
  };

  TacoscriptPrinter.prototype.printStatements = function printStatements(parent, prop) {
    var opts = arguments.length <= 2 || arguments[2] === undefined ? {} : arguments[2];

    opts.statement = true;
    return this.printMultiple(parent, prop, opts);
  };

  TacoscriptPrinter.prototype.printArguments = function printArguments(parent) {
    var _this2 = this;

    var prop = arguments.length <= 1 || arguments[1] === undefined ? "arguments" : arguments[1];
    var opts = arguments.length <= 2 || arguments[2] === undefined ? {} : arguments[2];

    var node = parent[prop];
    if (this.format.preserve && node.tokenElements && node.tokenElements.length) {
      throw new Error('Not Implemented');
    } else {
      (function () {
        // a transform will be used, so don't worry about exec (!) style arguments here.
        // reminder: exec style arguments are an arguments list that instead of being surrounded
        // by parens, starts with a !
        //
        // if exec style can't be inferred, <!-- will be generated at the end of the line
        // so, ! ...arguments #<!-- will be an illegal expression in tacoscript.
        _this2.push("(");
        opts.separator = ",";

        if (opts.indent) _this2.indent();

        var argNode = undefined,
            i = undefined,
            len = node.length;
        var after = function after() {
          if (opts.iterator) {
            opts.iterator(argNode, i);
          }
          if (opts.separator && i < len - 1) {
            // TODO: reenable this special case
            // if (!this.isLastType(tt.newline)) {
            _this2.push(",");
            // }
          }
        };

        var printOpts = {
          statement: opts.statement,
          after: after
        };

        for (i = 0; i < len; i++) {
          argNode = node[i];
          if (argNode) {
            _this2._simplePrint(argNode, parent, printOpts);
          } else {
            // preserve holes, especially in arrays
            _this2.push("pass");
            after();
          }
        }

        if (opts.indent) _this2.dedent();
        _this2.push(")");
      })();
    }
  };

  // for array and object literals
  // TODO: move to generators/taco/types

  TacoscriptPrinter.prototype.printLiteralBody = function printLiteralBody(parent, prop) {
    var opts = arguments.length <= 2 || arguments[2] === undefined ? {} : arguments[2];

    var node = parent[prop];
    if (this.format.preserve && parent.tokenElements && parent.tokenElements.length) {
      throw new Error('Not Implemented');
    } else {
      var useNewlines = true;
      if (parent.type === "ArrayExpression" && node.length <= 5) useNewlines = false;
      if (parent.type === "ArrayPattern") {
        useNewlines = false;
      }
      // if ((parent.type === "ObjectExpression" || parent.type === "ObjectPattern") && node.length <= 2) useNewlines = false;
      if (parent.type === "ObjectExpression" && node.length === 0) useNewlines = false;
      if (t.isObjectPattern(parent) && node.length <= 2) useNewlines = false;
      // TODO: implement the following line:
      //   if (parent.type === "ObjectPattern" && traversal.inLoopHead(parent)) useNewlines = false;
      // TODO: always use newlines if the literal contains a function
      opts.separator = useNewlines ? { type: "newline" } : ",";
      opts.indent = useNewlines;
      if (useNewlines) this.newline();
      this.printInnerComments(parent);
      this._simplePrintMultiple(node, parent, opts);
      if (useNewlines) this.newline();
    }
  };

  TacoscriptPrinter.prototype.printList = function printList(parent, prop) {
    var opts = arguments.length <= 2 || arguments[2] === undefined ? {} : arguments[2];

    if (opts.separator == null) {
      opts.separator = { type: ',' };
    }
    return this.printMultiple(parent, prop, opts);
    // TODO
    // eventually, don't print commas if a newline is available as a separator.
    // when source isn't available, preference should be supplied in opts:
    //   array should use newlines
    //   arguments should use commas
    //   also, there will be a format option to always insert commas
    // this will be passed as an argument to printMultiple
  };

  TacoscriptPrinter.prototype.printBlock = function printBlock(parent) {
    var prop = arguments.length <= 1 || arguments[1] === undefined ? 'body' : arguments[1];

    var node = parent[prop];
    var opts = {};
    // BlockStatement should only be printed with the generator when it is not
    // the body of a statement (such as if, etc.)
    if (t.isBlock(node)) {
      this.indent();
      this.newline();
      this._startPrint(parent, prop, opts);
      this.printStatements(node, 'directives', opts);
      this.printStatements(node, 'body', opts);
      this._finishPrint(node, parent, opts);
      this.dedent();
      // } else if (t.isEmptyStatement(node)) {
      //   // probably not needed
      //   this.push({type: 'pass', after: [';', '\n']}, parent, prop, opts);
    } else {
        // This is a single statement with no surrounding braces
        var noThen = t.isIfStatement(parent) && prop === 'alternate';
        noThen = noThen || t.isDoWhileStatement(parent) && prop === 'body';
        if (!noThen) this.push('then');
        this.print(parent, prop);
      }
  };

  TacoscriptPrinter.prototype.printComment = function printComment(comment) {
    if (!this.shouldPrintComment(comment)) return;

    if (comment.ignore) return;
    comment.ignore = true;

    if (comment.start != null) {
      if (this._printedCommentStarts[comment.start]) return;
      this._printedCommentStarts[comment.start] = true;
    }

    if (comment.type === "CommentBlock") {
      this._push({ type: _horchata.tokTypes.blockCommentStart, value: { kind: "#*", code: "#*" } });
      // TODO: encode/decode comment value for output in javascript
      var commentBody = comment.value;
      commentBody = commentBody.replace(blockCommentJs.terminatorEscapeSubRe, blockCommentJs.terminatorSub);
      commentBody = commentBody.replace(blockCommentTaco.terminatorSubRe, blockCommentTaco.terminatorEscapeSub);

      this._push({ type: _horchata.tokTypes.blockCommentBody, value: { kind: "#*", code: commentBody, value: comment.value } });
      this._push({ type: _horchata.tokTypes.blockCommentEnd, value: { kind: "#*", code: "*#" } });
    } else {
      var commentKind = undefined;switch (comment.end - comment.start - comment.value.length) {
        case 3:
          commentKind = "-->";break;
        case 4:
          commentKind = "<!--";break;
        default:
          commentKind = "#";
      }
      this._push({ type: _horchata.tokTypes.lineCommentStart, value: { kind: commentKind, code: commentKind } });
      this._push({ type: _horchata.tokTypes.lineCommentBody, value: { kind: commentKind, code: comment.value } });
      this.newline();
    }
  };

  TacoscriptPrinter.prototype.printLeadingComments = function printLeadingComments(node) {
    var comments = this.getComments("leadingComments", node);

    this.printComments(comments);

    // TOOD: convert to a "catchup" style newline printing
    // print newlines after leading comments
    var lastComment = comments[comments.length - 1];

    if (lastComment && lastComment.type === "CommentBlock" && node.loc.start.line > lastComment.loc.end.line) {
      this.newline();
    }
  };

  TacoscriptPrinter.prototype.printTrailingComments = function printTrailingComments(node) {
    var comments = this.getComments("trailingComments", node);
    this.printComments(comments);
  };

  return TacoscriptPrinter;
})(_tacoBuffer2["default"]);

exports["default"] = TacoscriptPrinter;

_Object$assign(TacoscriptPrinter.prototype, _printer3["default"]);

var _arr = [baseGenerators, classesGenerators, expressionsGenerators, literalsGenerators, methodsGenerators, modulesGenerators, statementsGenerators, templateLiteralsGenerators];

for (var _i = 0; _i < _arr.length; _i++) {
  var generator = _arr[_i];
  _Object$assign(TacoscriptPrinter.prototype, generator);
}
module.exports = exports["default"];